/*
 * Copyright (c) 2011 - 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <dlog.h>
#include <glib.h>
#include <db-util.h>
#include <shortcut_private.h>
#include <shortcut_db.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <tzplatform_config.h>

#include "shortcut_error.h"

#define QUERY_CREATE_SHORTCUT_TABLE \
	"PRAGMA journal_mode = PERSIST;" \
	"CREATE TABLE IF NOT EXISTS shortcut_service (\n" \
	"  id INTEGER PRIMARY KEY AUTOINCREMENT,\n" \
	"  pkgid TEXT,\n" \
	"  appid TEXT,\n" \
	"  icon TEXT,\n" \
	"  name TEXT,\n" \
	"  extra_key TEXT,\n" \
	"  extra_data TEXT);\n" \
	"CREATE TABLE IF NOT EXISTS shortcut_name (\n" \
	"  id INTEGER,\n" \
	"  pkgid TEXT,\n" \
	"  lang TEXT,\n" \
	"  name TEXT,\n" \
	"  icon TEXT);"

EAPI int shortcut_db_init()
{
	int ret = SHORTCUT_ERROR_NONE;
	int sql_ret;
	sqlite3 *db = NULL;
	char *errmsg = NULL;

	sql_ret = sqlite3_open_v2(DB_PATH, &db,
			SQLITE_OPEN_CREATE | SQLITE_OPEN_READWRITE, NULL);
	if (sql_ret != SQLITE_OK) {
		SHORTCUT_ERR("Failed to open db[%d]", sql_ret);
		return SHORTCUT_ERROR_IO_ERROR;
	}

	sql_ret = sqlite3_exec(db, QUERY_CREATE_SHORTCUT_TABLE,
					NULL, NULL, &errmsg);
	if (sql_ret != SQLITE_OK) {
		SHORTCUT_ERR("Failed to exec query [%d][%s]", sql_ret, errmsg);
		ret = SHORTCUT_ERROR_IO_ERROR;
	}

	if (errmsg)
		sqlite3_free(errmsg);
	if (db)
		sqlite3_close(db);

	return ret;
}

static sqlite3 *_open_db(void)
{
	int ret;
	const char *dbfile = DB_PATH;
	sqlite3 *db = NULL;

	ret = db_util_open(dbfile, &db, 0);
	if (ret != SQLITE_OK) {
		/* LCOV_EXCL_START */
		SHORTCUT_DBG("Failed to open a %s\n", dbfile);
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	return db;
}

static int _close_db(sqlite3 **db)
{
	int ret = 0;

	if (db == NULL || *db == NULL)
		return SHORTCUT_ERROR_INVALID_PARAMETER;

	ret = db_util_close(*db);
	if (ret != SQLITE_OK) {
		/* LCOV_EXCL_START */
		SHORTCUT_DBG("DB close error(%d)", ret);
		return SHORTCUT_ERROR_IO_ERROR;
		/* LCOV_EXCL_STOP */
	}

	*db = NULL;

	return SHORTCUT_ERROR_NONE;
}

/* LCOV_EXCL_START */
/*!
 * \note this function will returns allocated(heap) string
 */
static inline int _get_i18n_name(sqlite3 *handle, const char *lang, int id, char **name, char **icon)
{
	sqlite3_stmt *stmt;
	static const char *query = "SELECT name, icon FROM shortcut_name WHERE id = ? AND lang = ? COLLATE NOCASE";
	const unsigned char *_name;
	const unsigned char *_icon;
	int ret = 0;
	int status;

	status = sqlite3_prepare_v2(handle, query, -1, &stmt, NULL);
	if (status != SQLITE_OK) {
		SHORTCUT_ERR("Failed to prepare stmt: %s\n", sqlite3_errmsg(handle));
		return -EFAULT;
	}

	status = sqlite3_bind_int(stmt, 1, id);
	if (status != SQLITE_OK) {
		SHORTCUT_ERR("Failed to bind id: %s\n", sqlite3_errmsg(handle));
		ret = -EFAULT;
		goto out;
	}

	status = sqlite3_bind_text(stmt, 2, lang, -1, SQLITE_TRANSIENT);
	if (status != SQLITE_OK) {
		SHORTCUT_ERR("Failed to bind lang: %s\n", sqlite3_errmsg(handle));
		ret = -EFAULT;
		goto out;
	}

	SHORTCUT_DBG("id: %d, lang: %s\n", id, lang);
	if (SQLITE_ROW != sqlite3_step(stmt)) {
		SHORTCUT_ERR("Failed to do step: %s\n", sqlite3_errmsg(handle));
		ret = -ENOENT;
		goto out;
	}

	_name = sqlite3_column_text(stmt, 0);
	if (name) {
		if (_name && strlen((const char *)_name)) {
			*name = strdup((const char *)_name);
			if (!*name) {
				SHORTCUT_ERR("strdup: %d\n", errno);
				ret = -ENOMEM;
				goto out;
			}
		} else {
			*name = NULL;
		}
	}

	_icon = sqlite3_column_text(stmt, 1);
	if (icon) {
		if (_icon && strlen((const char *)_icon)) {
			*icon = strdup((const char *)_icon);
			if (!*icon) {
				SHORTCUT_ERR("strdup: %d\n", errno);
				ret = -ENOMEM;
				if (name && *name)
					free(*name);

				goto out;
			}
		} else {
			*icon = NULL;
		}
	}

out:
	sqlite3_reset(stmt);
	sqlite3_clear_bindings(stmt);
	sqlite3_finalize(stmt);
	return ret;
}
/* LCOV_EXCL_STOP */

static inline char *_cur_locale(void)
{
	char *language;
	char *ptr;

	language = vconf_get_str(VCONFKEY_LANGSET);
	if (language) {
		ptr = language;
		while (*ptr) {
			if (*ptr == '.') {
				*ptr = '\0';
				break;
			}

			if (*ptr == '_')
				*ptr = '-';

			ptr++;
		}
	} else {
		language = strdup("en-us");
		if (!language)
			SHORTCUT_ERR("Heap: %d\n", errno);
	}

	return language;
}

EAPI int shortcut_db_get_list(const char *package_name, GList **shortcut_list)
{
	sqlite3_stmt *stmt;
	sqlite3 *handle = NULL;
	const char *query;
	const unsigned char *name;
	char *i18n_name = NULL;
	char *i18n_icon = NULL;
	const unsigned char *extra_data;
	const unsigned char *extra_key;
	const unsigned char *icon;
	shortcut_info_s *shortcut;
	int id;
	int ret;
	int cnt;
	char *language;

	handle = _open_db();
	if (!handle) {
		/* LCOV_EXCL_START */
		SHORTCUT_ERR("Failed to open a DB\n");
		return SHORTCUT_ERROR_IO_ERROR;
		/* LCOV_EXCL_STOP */
	}

	language = _cur_locale();
	if (!language) {
		/* LCOV_EXCL_START */
		SHORTCUT_ERR("Locale is not valid\n");
		_close_db(&handle);
		return SHORTCUT_ERROR_FAULT;
		/* LCOV_EXCL_STOP */
	}

	if (package_name) {
		query = "SELECT id, appid, name, extra_key, extra_data, icon FROM shortcut_service WHERE appid = ?";
		ret = sqlite3_prepare_v2(handle, query, -1, &stmt, NULL);
		if (ret != SQLITE_OK) {
			/* LCOV_EXCL_START */
			SHORTCUT_ERR("prepare: %s\n", sqlite3_errmsg(handle));
			free(language);
			_close_db(&handle);
			return SHORTCUT_ERROR_IO_ERROR;
			/* LCOV_EXCL_STOP */
		}

		ret = sqlite3_bind_text(stmt, 1, package_name, -1, SQLITE_TRANSIENT);
		if (ret != SQLITE_OK) {
			/* LCOV_EXCL_START */
			SHORTCUT_ERR("bind text: %s\n", sqlite3_errmsg(handle));
			sqlite3_finalize(stmt);
			free(language);
			_close_db(&handle);
			return SHORTCUT_ERROR_IO_ERROR;
			/* LCOV_EXCL_STOP */
		}
	} else {
		query = "SELECT id, appid, name, extra_key, extra_data, icon FROM shortcut_service";
		ret = sqlite3_prepare_v2(handle, query, -1, &stmt, NULL);
		if (ret != SQLITE_OK) {
			/* LCOV_EXCL_START */
			SHORTCUT_ERR("prepare: %s\n", sqlite3_errmsg(handle));
			free(language);
			_close_db(&handle);
			return SHORTCUT_ERROR_IO_ERROR;
			/* LCOV_EXCL_STOP */
		}
	}

	cnt = 0;
	*shortcut_list = NULL;
	while (SQLITE_ROW == sqlite3_step(stmt)) {
		/* LCOV_EXCL_START */
		id = sqlite3_column_int(stmt, 0);

		package_name = (const char *)sqlite3_column_text(stmt, 1);
		if (!package_name) {
			LOGE("Failed to get package name\n");
			continue;
		}

		name = sqlite3_column_text(stmt, 2);
		if (!name) {
			LOGE("Failed to get name\n");
			continue;
		}

		extra_key = sqlite3_column_text(stmt, 3);
		if (!extra_key) {
			LOGE("Failed to get service\n");
			continue;
		}

		extra_data = sqlite3_column_text(stmt, 4);
		if (!extra_data) {
			LOGE("Failed to get service\n");
			continue;
		}

		icon = sqlite3_column_text(stmt, 5);
		if (!icon) {
			LOGE("Failed to get icon\n");
			continue;
		}

		/*!
		 * \todo
		 * Implement the "GET LOCALE" code
		 */
		/* if (get_i18n_name(language, id, &i18n_name, &i18n_icon) < 0) { */
			/* Okay, we can't manage this. just use the fallback string */
		/* } */
		_get_i18n_name(handle, language, id, &i18n_name, &i18n_icon);

		cnt++;
		shortcut = (shortcut_info_s *)calloc(sizeof(shortcut_info_s), 1);
		if (shortcut == NULL) {
			free(i18n_name);
			break;
		}
		shortcut->package_name = strdup(package_name);
		shortcut->icon = strdup((i18n_icon != NULL ? i18n_icon : (char *)icon));
		shortcut->name = strdup((i18n_name != NULL ? i18n_name : (char *)name));
		shortcut->extra_key = strdup((char *)extra_key);
		shortcut->extra_data = strdup((char *)extra_data);
		*shortcut_list = g_list_append(*shortcut_list, shortcut);

		free(i18n_name);
		i18n_name = NULL;

		free(i18n_icon);
		i18n_icon = NULL;
		/* LCOV_EXCL_STOP */
	}

	sqlite3_reset(stmt);
	sqlite3_clear_bindings(stmt);
	sqlite3_finalize(stmt);
	free(language);
	_close_db(&handle);

	return cnt;
}
